using System;
using System.Net;
using System.Xml;
using System.Xml.Schema;
using pico.ap;
using pico.oaipmh2;
using System.Collections;
using System.IO;
using System.Configuration;

//using org.apache.xerces.parsers;
//using org.xml.sax;

namespace pico.validator
{  
	/// <summary>
	/// Descrizione di riepilogo per Class1.
	/// </summary>
	class Class1
	{
    static private string _metadataPrefix = "pico";
    static private Harvester _harvester;    
    static private ApplicationProfile _apPICO = null;
    static private ApplicationProfile _apMAG = null;
    static private ApplicationProfile _apICCD = null;
    static private Validator _validator = null;
    static private string _currSet;
    static private XmlDocument _thesaurus;
    static private XmlDocument _thesaurus4;
    static private XmlNamespaceManager _thesaurusNamespaceManager;
    static private XmlNamespaceManager _thesaurus4NamespaceManager;
    static private int _thesSubjectCount = 0;
    static private int _thes4SubjectCount = 0;
    static private int _subjectCount = 0;
    static private int _titleCount = 0;
    static private int _typeCount = 0;
    static private int _identifierCount = 0;
    static private string _setFilter = null;
    static private string _inputFile = null;
    static private int _httpRequestTimeout = 5000;

    static private Hashtable _setStatInfo = new Hashtable();

		/// <summary>
		/// Il punto di ingresso principale dell'applicazione.
		/// </summary>
		[STAThread]
		static void Main(string[] args)
		{
			CheckArguments(args);

      CheckAppSettings();

      string repositoryAddress = args[0];

      try
      {
        new Uri(repositoryAddress);

        if(!repositoryAddress.StartsWith("http://"))
          throw new UriFormatException("URI non valido: prefisso sconosciuto");
      }
      catch(UriFormatException e)
      {
        System.Console.Error.WriteLine(e.Message);
        System.Environment.Exit(-2);
      }

      if(args.Length == 3)
      {
        char op = args[1][1];

        if(op == 's')
          _setFilter = args[2];
        else if(op == 'i')
          _inputFile = args[2];        
      }

      if(ConfigurationSettings.AppSettings["OaiPmhMetadataPrefix"] != null)
        _metadataPrefix = ConfigurationSettings.AppSettings["OaiPmhMetadataPrefix"];

      PrintVersion();

      _httpRequestTimeout = int.Parse(ConfigurationSettings.AppSettings["HttpRequestTimeout"]);

      // Carica il thesaurus pico 3.x
      string picoThes3Url = ConfigurationSettings.AppSettings["PicoThes3Url"];
      _thesaurus = LoadXmlDocument(picoThes3Url);
      if(_thesaurus != null)
        System.Console.Out.WriteLine("Caricato thesaurus 3.x: {0}", picoThes3Url);
      else
        System.Environment.Exit(-3);

      // Carica il thesaurus pico 4.x
      string picoThes4Url = ConfigurationSettings.AppSettings["PicoThes4Url"];
      _thesaurus4 = LoadXmlDocument(picoThes4Url);
      if(_thesaurus4 != null)
        System.Console.Out.WriteLine("Caricato thesaurus 4.x: {0}", picoThes4Url);
      else
        System.Environment.Exit(-4);

      // Carica il PICO AP
      string picoApUrl = ConfigurationSettings.AppSettings["PicoApUrl"];
      XmlDocument xmlDocPICOAP = LoadXmlDocument(picoApUrl);
      System.Console.Out.WriteLine("Caricato profilo applicativo: {0}", picoApUrl);
      _apPICO = new ApplicationProfile(xmlDocPICOAP);
      
      // Carica il sottoprofilo applicativo MAG
      string picoSubApMagUrl = ConfigurationSettings.AppSettings["PicoSubApMagUrl"];
      XmlDocument xmlDocMAGAP = LoadXmlDocument(picoSubApMagUrl);
      System.Console.Out.WriteLine("Caricato profilo applicativo: {0}", picoSubApMagUrl);
      _apMAG = new ApplicationProfile(xmlDocMAGAP);

      // Aggiunge il sottoprofilo applicativo MAG al PICOAP
      _apPICO.Add(_apMAG);

      // Carica il sottoprofilo applicativo ICCD
      _apICCD = LoadSubProfilesICCD();

      // Aggiunge il sottoprofilo applicativo ICCD al PICOAP
      _apPICO.Add(_apICCD);
            
      string errorLevel = ConfigurationSettings.AppSettings["ErrorLevel"];
      // Crea l'oggetto Validator
      _validator = new Validator(_apPICO);
      _validator.SetErrorLevel(errorLevel[0]);

      // Crea il namespace manager per il thesaurus 3.x
      _thesaurusNamespaceManager = new System.Xml.XmlNamespaceManager(_thesaurus.NameTable);
      //Aggiunge i namespace del thesaurus
      _thesaurusNamespaceManager.AddNamespace("th", "http://pico.sns.it/pico/thesaurus/");
      _thesaurusNamespaceManager.AddNamespace("dc", "http://purl.org/dc/elements/1.1/");
      _thesaurusNamespaceManager.AddNamespace("dcterms", "http://purl.org/dc/terms/");
      _thesaurusNamespaceManager.AddNamespace("xsi", "http://www.w3.org/2001/XMLSchema-instance");

      // Crea il namespace manager per il thesaurus 4.x
      _thesaurus4NamespaceManager = new System.Xml.XmlNamespaceManager(_thesaurus4.NameTable);
      //Aggiunge i namespace del thesaurus
      _thesaurus4NamespaceManager.AddNamespace("rdf", "http://www.w3.org/1999/02/22-rdf-syntax-ns#");
      _thesaurus4NamespaceManager.AddNamespace("skos", "http://www.w3.org/2004/02/skos/core#");
		  _thesaurus4NamespaceManager.AddNamespace("dc", "http://purl.org/dc/elements/1.1/");
		  _thesaurus4NamespaceManager.AddNamespace("dcterms", "http://purl.org/dc/terms/");

      // Imposta il gestore degli errori del validatore
      _validator.SetErrorHandler(new Validator.ErrorHandler(ProcessValidatorError));

      // Imposta il gestore dei metadati
      _validator.SetMetadataHandler(new Validator.MetadataHandler(ProcessMetadata));

      // Crea l'harvester
      _harvester = new Harvester();
      
      // Importa l'uri di base dell'harvester
      _harvester.SetBaseUri(repositoryAddress);

      // Imposta il gestore dei set
      _harvester.SetSetHandler(new Harvester.SetHandler(ProcessSet));
            
      // Imposta il gestore degli identifier
      _harvester.SetIdentifierHandler(new Harvester.IdentifierHandler(ProcessIdentifier));

      // Imposta il gestore dei record
      _harvester.SetRecordHandler(new Harvester.RecordHandler(ProcessRecord));
      
      // Imposta il gestore degli errori http
      _harvester.SetHttpErrorHandler(new Harvester.HttpErrorHandler(ProcessHttpError));

      // Imposta il gestore degli errori dell'harvester
      _harvester.SetErrorHandler(new Harvester.ErrorHandler(ProcessHarvesterError));
      
      // Imposta vari parametri per il funzionamento
      _harvester.ListIdentifiersDelay = int.Parse(ConfigurationSettings.AppSettings["HarvesterListIdentifiersDelay"]);

      _harvester.HttpRequestTimeout = int.Parse(ConfigurationSettings.AppSettings["HarvesterHttpRequestTimeout"]);

      _harvester.UserAgent = ConfigurationSettings.AppSettings["HarvesterUserAgent"];

      _harvester.MailFrom =  ConfigurationSettings.AppSettings["HarvesterMailFrom"];

      if(_inputFile == null)
      {
        // Esegue un comando ListSets
        _harvester.ListSets(null);

        // Se il repository non espone nessun set esegue una 
        // ListIdentifiers senza specifica del set
        if(_harvester.GetListedSetsCount() == 0)
        {
          _setStatInfo.Add("NULL", new SetStatInfo());
          _harvester.ListIdentifiers(null, "pico", null);
        }
      }
      else
      {
        _currSet = _inputFile;
        _setStatInfo.Add(_inputFile, new SetStatInfo());
        ReadInputFile();
      }
      
      // Stampa le statistiche di harvesting
      PrintStatInfo();
		}

    static void ReadInputFile()
    {
      if(!File.Exists(_inputFile))
        return;

      StreamReader sr = new StreamReader(_inputFile);
      string line = null;
     
      while((line = sr.ReadLine()) != null)
      {
        ProcessIdentifier(_inputFile, line);

        XmlDocument xmlRec = _harvester.GetRecord(line, _metadataPrefix);

        if(xmlRec != null)
        {
          ProcessRecord(_inputFile, line, xmlRec);
        }
      }

      sr.Close();
    }

    // Helper di Main()
    static void CheckArguments(string[] args)
    {
      bool exit = false;

      if(args.Length < 1 || args.Length > 5)
        exit = true;
      
      if(args.Length == 3 || args.Length == 5)
      {
        for(int n = 1; n < args.Length; n+=2)
        {
          if(!args[n].Equals("-s") && !args[n].Equals("-i"))
          {
            exit = true;
            break;
          }
        }
      }
      else if(args.Length != 1)
      {
        exit = true;
      }

      if(exit)
      {
        Usage();
        System.Environment.Exit(-1);
      }
    }

    static void CheckAppSettings()
    {      
      string[] keyNames = new string[] {
        "HttpRequestTimeout",
        "HarvesterHttpRequestTimeout",
        "HarvesterListIdentifiersDelay",
        "HarvesterUserAgent",
        "HarvesterMailFrom",
        "PicoThes3Url",
        "PicoThes4Url",
        "PicoApUrl",
        "PicoSubApIccdUrl",
        "PicoSubApIccdOadnUrl",
        "PicoSubApIccdRaUrl",
        "PicoSubApIccdSmiUrl",
        "PicoSubApIccdBdmUrl",
        "PicoSubApIccdFUrl",
        "PicoSubApIccdBibUrl",
        "PicoSubApIccd3Url",
        "PicoSubApIccdNu3Url",
        "PicoSubApIccdF3Url",
        "PicoSubApMagUrl",
        "ErrorLevel",
        //"OaiPmhMetadataPrefix" non  abbligatorio perche' il default  "pico"
        };

      foreach(string keyName in keyNames)
      {
        if(ConfigurationSettings.AppSettings[keyName] == null)
        {
          System.Console.Error.WriteLine("File di configurazione mancante o corrotto.");
          System.Environment.Exit(-1);
        }
      }
    }

    static void PrintVersion()
    {
      string appName = System.Environment.CommandLine.Split(' ')[0];

      string version = System.Reflection.Assembly.GetExecutingAssembly().GetName().Version.ToString();

      System.Console.Out.WriteLine("{0} v. {1}", appName, version);
    }

    // Helper di Main()
    static void Usage()
    {
      string appName = System.Environment.CommandLine.Split(' ')[0];

      System.Console.Out.WriteLine("Utilizzo:");
      System.Console.Out.WriteLine("{0} repository_address_url [-s set_name] [-i input_file]",
        appName);
      System.Console.Out.WriteLine("Esempi:");
      System.Console.Out.WriteLine("validate http://192.168.1.1/OAIHandler");
      System.Console.Out.WriteLine("validate http://192.168.1.1/OAIHandler -s autori");
      System.Console.Out.WriteLine("validate http://192.168.1.1/OAIHandler -i ids.txt");      
    }

    // Helper di Main()
    // Carica un documento xml tramite url con protocollo http o tramite path
    // assoluto o relativo all'applicazione
    static XmlDocument LoadXmlDocument(string url)
    {
      XmlDocument xmlDoc = null;

      if(!url.StartsWith(Uri.UriSchemeHttp))
      {
        string fullPath = Path.IsPathRooted(url) ? url : Path.Combine(System.Environment.CurrentDirectory, url);
        url = string.Format("{0}://{1}", Uri.UriSchemeFile, fullPath);

        XmlTextReader xmlTextReader = new XmlTextReader(url);
      
        try
        {
          xmlDoc = new XmlDocument();        
          xmlDoc.Load(xmlTextReader);
        }
        catch(Exception e)
        {
          xmlDoc = null;
          System.Console.Error.WriteLine("Impossibile leggere il documento xml:\n'{0}\n{1}",
            url, e.Message);
        }
      }
      else
      {
        HttpWebRequest req = (HttpWebRequest)WebRequest.Create(url);
        req.Timeout = _httpRequestTimeout; //int.Parse(ConfigurationSettings.AppSettings["HttpRequestTimeout"]);

        string xml = "";

        try
        {
          HttpWebResponse res = (HttpWebResponse)req.GetResponse();

          res.GetResponseStream();

          HttpStatusCode httpStatusCode = res.StatusCode;

          if(httpStatusCode == HttpStatusCode.OK)
          {
            xml = ReadContent(res);

            xmlDoc = new XmlDocument();        

            xmlDoc.LoadXml(xml);
          }
        }
        catch(WebException webExc)
        {
          System.Console.Error.WriteLine("Impossibile leggere il documento xml:\n'{0}'\n{1}",
            url, webExc.Message);
        }
      }
      
      return xmlDoc;
    }

    // Helper di LoadXmlDocument()
    static string ReadContent(HttpWebResponse res)
    {
      int bytesRead = 0;
      char[] buffer = new char[512];

      StreamReader reader = new StreamReader(res.GetResponseStream(), System.Text.Encoding.UTF8);
      StringWriter writer = new StringWriter();

      bytesRead = reader.Read(buffer, 0, 512);
      while (bytesRead != 0 ) 
      {
        writer.Write(buffer, 0, bytesRead);
        bytesRead = reader.Read(buffer, 0, 512);
      }

      return writer.ToString();
    }

    // Helper di Main()
    static ApplicationProfile LoadSubProfilesICCD()
    {
      string picoSubApIccd = ConfigurationSettings.AppSettings["PicoSubApIccdUrl"];
      XmlDocument xmlDocICCD = LoadXmlDocument(picoSubApIccd);
      System.Console.Out.WriteLine("Caricato profilo applicativo: {0}", picoSubApIccd);
      ApplicationProfile apICCD = new ApplicationProfile(xmlDocICCD);

      string picoSubApIccdRa = ConfigurationSettings.AppSettings["PicoSubApIccdRaUrl"];
      XmlDocument xmlDocRA = LoadXmlDocument(picoSubApIccdRa);
      System.Console.Out.WriteLine("Caricato profilo applicativo: {0}", picoSubApIccdRa);
      ApplicationProfile apRA = new ApplicationProfile(xmlDocRA);
      apICCD.Add(apRA);

      string picoSubApIccdOadn = ConfigurationSettings.AppSettings["PicoSubApIccdOadnUrl"];
      XmlDocument xmlDocOADN = LoadXmlDocument(picoSubApIccdOadn);
      System.Console.Out.WriteLine("Caricato profilo applicativo: {0}", picoSubApIccdOadn);
      ApplicationProfile apOADN = new ApplicationProfile(xmlDocOADN);
      apICCD.Add(apOADN);

      string picoSubApIccdSmi = ConfigurationSettings.AppSettings["PicoSubApIccdSmiUrl"];
      XmlDocument xmlDocSMI = LoadXmlDocument(picoSubApIccdSmi);
      System.Console.Out.WriteLine("Caricato profilo applicativo: {0}", picoSubApIccdSmi);
      ApplicationProfile apSMI = new ApplicationProfile(xmlDocSMI);
      apICCD.Add(apSMI);

      string picoSubApIccdBdm = ConfigurationSettings.AppSettings["PicoSubApIccdBdmUrl"];
      XmlDocument xmlDocBDM = LoadXmlDocument(picoSubApIccdBdm);
      System.Console.Out.WriteLine("Caricato profilo applicativo: {0}", picoSubApIccdBdm);
      ApplicationProfile apBDM = new ApplicationProfile(xmlDocBDM);
      apICCD.Add(apBDM);

      string picoSubApIccdF = ConfigurationSettings.AppSettings["PicoSubApIccdFUrl"];
      XmlDocument xmlDocF = LoadXmlDocument(picoSubApIccdF);
      System.Console.Out.WriteLine("Caricato profilo applicativo: {0}", picoSubApIccdF);
      ApplicationProfile apF = new ApplicationProfile(xmlDocF);
      apICCD.Add(apF);

      string picoSubApIccdBib = ConfigurationSettings.AppSettings["PicoSubApIccdBibUrl"];
      XmlDocument xmlDocBIB = LoadXmlDocument(picoSubApIccdBib);
      System.Console.Out.WriteLine("Caricato profilo applicativo: {0}", picoSubApIccdBib);
      ApplicationProfile apBIB = new ApplicationProfile(xmlDocBIB);
      apICCD.Add(apBIB);

      string picoSubApIccd3 = ConfigurationSettings.AppSettings["PicoSubApIccd3Url"];
      XmlDocument xmlDocICCD3 = LoadXmlDocument(picoSubApIccd3);
      System.Console.Out.WriteLine("Caricato profilo applicativo: {0}", picoSubApIccd3);
      ApplicationProfile apICCD3 = new ApplicationProfile(xmlDocICCD3);
      apICCD.Add(apICCD3);

      string picoSubApIccd3Nu = ConfigurationSettings.AppSettings["PicoSubApIccdNu3Url"];
      XmlDocument xmlDocNU3 = LoadXmlDocument(picoSubApIccd3Nu);
      System.Console.Out.WriteLine("Caricato profilo applicativo: {0}", picoSubApIccd3Nu);
      ApplicationProfile apNU3 = new ApplicationProfile(xmlDocNU3);
      apICCD.Add(apNU3);

      string picoSubApIccd3F = ConfigurationSettings.AppSettings["PicoSubApIccdF3Url"];
      XmlDocument xmlDocF3 = LoadXmlDocument(picoSubApIccd3F);
      System.Console.Out.WriteLine("Caricato profilo applicativo: {0}", picoSubApIccd3F);
      ApplicationProfile apF3 = new ApplicationProfile(xmlDocF3);
      apICCD.Add(apF3);

      return apICCD;
    }

    /* Metodo temporaneo che serve per caricare manualmente il sotto-profilo iccd in attesa
     * di avere il sotto-profilo xml on line.
     * */
    static void LoadSubProfiles(ref ApplicationProfile ap)
    {
      ApplicationProfile.TermUsage tuCD = new ApplicationProfile.TermUsage();
      tuCD.Type = ApplicationProfile.TermUsage.TermUsageType.EncodingScheme;
      tuCD.DataType = ApplicationProfile.TermUsage.TermUsageDataType.String;
      tuCD.Name = "CD";
      tuCD.Uri = "http://pico.sns.it/pico/iccd/3.00/CD";
      ap.AddTermUsage(tuCD);

      ApplicationProfile.TermUsage tuNCT = new ApplicationProfile.TermUsage();
      tuNCT.Type = ApplicationProfile.TermUsage.TermUsageType.EncodingScheme;
      tuNCT.DataType = ApplicationProfile.TermUsage.TermUsageDataType.String;
      tuNCT.Name = "NCT";
      tuNCT.Uri = "http://pico.sns.it/pico/iccd/3.00/NCT";
      ap.AddTermUsage(tuNCT);

      ApplicationProfile.TermUsage tuOGT = new ApplicationProfile.TermUsage();
      tuOGT.Type = ApplicationProfile.TermUsage.TermUsageType.EncodingScheme;
      tuOGT.DataType = ApplicationProfile.TermUsage.TermUsageDataType.String;
      tuOGT.Name = "OGT";
      tuOGT.Uri = "http://pico.sns.it/pico/iccd/3.00/f/OGT";
      ap.AddTermUsage(tuOGT);

      ApplicationProfile.TermUsage tuSGT = new ApplicationProfile.TermUsage();
      tuSGT.Type = ApplicationProfile.TermUsage.TermUsageType.EncodingScheme;
      tuSGT.DataType = ApplicationProfile.TermUsage.TermUsageDataType.String;
      tuSGT.Name = "SGT";
      tuSGT.Uri = "http://pico.sns.it/pico/iccd/3.00/f/SGT";
      ap.AddTermUsage(tuSGT);

      ApplicationProfile.TermUsage tuSGL = new ApplicationProfile.TermUsage();
      tuSGL.Type = ApplicationProfile.TermUsage.TermUsageType.EncodingScheme;
      tuSGL.DataType = ApplicationProfile.TermUsage.TermUsageDataType.String;
      tuSGL.Name = "SGL";
      tuSGL.Uri = "http://pico.sns.it/pico/iccd/3.00/f/SGL";
      ap.AddTermUsage(tuSGL);

      ApplicationProfile.TermUsage tuLR = new ApplicationProfile.TermUsage();
      tuLR.Type = ApplicationProfile.TermUsage.TermUsageType.EncodingScheme;
      tuLR.DataType = ApplicationProfile.TermUsage.TermUsageDataType.String;
      tuLR.Name = "LR";
      tuLR.Uri = "http://pico.sns.it/pico/iccd/3.00/f/LR";
      ap.AddTermUsage(tuLR);

      ApplicationProfile.TermUsage tuType = ap.FindTerm("type", "http://purl.org/dc/elements/1.1/");
      tuType.AddEncodingScheme(tuCD.Uri);
      tuType.AddEncodingScheme(tuOGT.Uri);

      ApplicationProfile.TermUsage tuTitle = ap.FindTerm("title", "http://purl.org/dc/elements/1.1/");
      tuTitle.AddEncodingScheme(tuSGL.Uri);

      ApplicationProfile.TermUsage tuIdentifier = ap.FindTerm("identifier", "http://purl.org/dc/elements/1.1/");
      tuIdentifier.AddEncodingScheme(tuNCT.Uri);

      ApplicationProfile.TermUsage tuSubject = ap.FindTerm("subject", "http://purl.org/dc/elements/1.1/");
      tuSubject.AddEncodingScheme(tuSGT.Uri);

      ApplicationProfile.TermUsage tuSpatial = ap.FindTerm("spatial", "http://purl.org/dc/terms/");
      tuSpatial.AddEncodingScheme(tuLR.Uri);
    }

    /// <summary>
    /// Il metodo che si occupa di gestire ogni singolo set harvestato
    /// Opera come delegato Harvester.SetHandler(...)
    /// </summary>
    /// <param name="setSpec">specifica del set da processare</param>
    /// <param name="setInfo">nodo xml contenente le informazioni relative al set</param>
    static void ProcessSet(string setSpec, XmlNode setInfo)
    {
      _currSet = setSpec;

      if(!_setStatInfo.ContainsKey(setSpec))
      {
        _setStatInfo.Add(setSpec, new SetStatInfo());
      }

      if(_setFilter == null)
        _harvester.ListIdentifiers(setSpec, _metadataPrefix, null);
      else
      {
        if(_setFilter.EndsWith(":"))
        {
          if(setSpec.StartsWith(_setFilter))
            _harvester.ListIdentifiers(setSpec, _metadataPrefix, null);
        }
        else
        {
          if(setSpec.Equals(_setFilter))
            _harvester.ListIdentifiers(setSpec, _metadataPrefix, null);
        }      
      }
    }

    static void ProcessIdentifier(string setSpec, string identifier)
    {
      if(setSpec == null)
        setSpec = "NULL";
      if(_setStatInfo.ContainsKey(setSpec))
      {
        ((SetStatInfo)_setStatInfo[setSpec]).totIdentifiers++;
        ((SetStatInfo)_setStatInfo[setSpec]).processed = true;
      }      
    }

    /// <summary>
    /// Il metodo che si occupa di gestire ogni singolo record harvestato
    /// </summary>
    /// <param name="identifier">identificatore del record da processare</param>
    /// <param name="rec">documento xml contenente il record da processare</param>
    static void ProcessRecord(string setSpec, string identifier, XmlDocument rec)
    {
      System.Console.Out.WriteLine(identifier);

      //CheckMandatoryFields();

      if(setSpec == null)
        setSpec = "NULL";
      if(_setStatInfo.ContainsKey(setSpec))
      {
        ((SetStatInfo)_setStatInfo[setSpec]).totRecords++;
      }

      XmlNode picoRec = FindNode(rec.DocumentElement, "pico:record");
      if(picoRec != null)
      {
        //bool isValid = SchemaValidateNode(picoRec);
        //if(isValid)

        XmlAttributeCollection attributes = picoRec.Attributes;
        foreach(XmlAttribute attr in attributes)
        {
          string prefix = attr.Prefix;

          if(prefix == "xmlns")
          {
            string uri = attr.Value;
            string pref = attr.LocalName;
            _validator.SetNamespace(pref, uri);
          }
        }

        _validator.Validate(picoRec);

        CheckMandatoryFields();
      }
    }

    static void CheckMandatoryFields()
    {
      if(_thesSubjectCount == 0 && _thes4SubjectCount == 0)
      {
        System.Console.Out.WriteLine("Attenzione: subject con encoding pico:Thesaurus mancante");
      }

      if(_subjectCount == 0)
      {
        System.Console.Out.WriteLine("Errore: subject mancante");
      }

      if(_titleCount == 0)
      {
        System.Console.Out.WriteLine("Errore: title mancante");
      }

      if(_typeCount == 0)
      {
        System.Console.Out.WriteLine("Errore: type mancante");
      }

      if(_identifierCount == 0)
      {
        System.Console.Out.WriteLine("Errore: identifier mancante");
      }

      _thesSubjectCount = 0;
      _thes4SubjectCount = 0;
      _subjectCount = 0;
      _titleCount = 0;
      _identifierCount = 0;
      _typeCount = 0;
    }

    static bool SchemaValidateNode2(XmlDocument doc)
    {
      /*
      SAXParser parser = new SAXParser();
      parser.setFeature("http://xml.org/sax/features/validation", true);
      parser.setFeature("http://apache.org/xml/features/validation/schema", true);
      parser.setFeature("http://apache.org/xml/features/validation/schema-full-checking", true);
      */

      return false;
    }

    static bool SchemaValidateNode(XmlNode node)
    {
      XmlNameTable nt = node.OwnerDocument.NameTable;
      XmlNamespaceManager nsMgr = new XmlNamespaceManager(nt);
      XmlSchemaCollection schemaCollection = new XmlSchemaCollection();

      foreach(XmlAttribute attr in node.Attributes)
      {
        if(attr.Prefix.Equals("xmlns"))
        {
          string name = attr.LocalName;
          string nsUri = attr.Value;

          nsMgr.AddNamespace(name, nsUri);
        }
        if(attr.Name.Equals("xsi:schemaLocation"))
        {
          string tmp = attr.Value;
          string val = "";
          StringReader stringReader = new StringReader(tmp);
          while(true)
          {
            string line = stringReader.ReadLine();
            if(line != null)
            {
              string []parts = line.Split((char)32);
              foreach(string part in parts)
              {
                if(part.Length > 0)
                {
                  if(val.Length > 0)
                    val = val += " ";
                  val = val += part;
                }
              }
            }
            else
              break;
          }

          char []seps = new char[] {' '};
          string []arr = val.Split(seps);

          if(arr.Length % 2 == 0)
          {
            for(int n = 0; n < arr.Length; n+=2)
            {
              string nsUri = arr[n];
              string xsdUrl = arr[n + 1];

              XmlReader schemaReader = new XmlTextReader(xsdUrl);
              schemaCollection.Add(nsUri, schemaReader);
            }
          }
        }
      }

      // Gli schema xsd inclusi tramite l'istruzione import da altri xsd non
      // vengono caricati automaticamente quindi occorre inserire a mano gli schema
      // dc.xsd e dcterms.xsd
      string nsdc = "http://purl.org/dc/elements/1.1/";      
      string nsdcterms = "http://purl.org/dc/terms/";      

      if(!schemaCollection.Contains(nsdc))
      {
        string urldc = "http://www.dublincore.org/schemas/xmls/qdc/dc.xsd";

        try
        {
          XmlReader schemaReader = new XmlTextReader(urldc);
          schemaCollection.Add(nsdc, schemaReader);
        }
        catch(Exception exc)
        {
          System.Console.Out.WriteLine(exc.Message);
        }
      }

      if(!schemaCollection.Contains(nsdcterms))
      {        
        string urldcterms = "http://www.dublincore.org/schemas/xmls/qdc/dcterms.xsd";
        try
        {
          XmlReader schemaReader = new XmlTextReader(urldcterms);
          schemaCollection.Add(nsdcterms, schemaReader);
        }
        catch(Exception exc)
        {
          System.Console.Out.WriteLine(exc.Message);
        }
      }

      XmlParserContext parserContext = new XmlParserContext(nt, nsMgr, "", XmlSpace.None);
      XmlValidatingReader validatingReader = new XmlValidatingReader(node.InnerXml, XmlNodeType.Element, parserContext);
      validatingReader.ValidationType = ValidationType.Schema;
      validatingReader.ValidationEventHandler += new ValidationEventHandler(MyValidationEventHandler);
      validatingReader.Schemas.Add(schemaCollection);
      
      try
      {
        while (validatingReader.Read())
        {
          //System.Console.Out.WriteLine(validatingReader.Name);
          // Can add code here to process the content.
        }
      }
      catch(Exception e)
      {
        System.Console.Out.WriteLine(e.Message);
      }

      validatingReader.Close();
      return true;
    }

    static bool SchemaValidateDocument(XmlDocument doc)
    {
      try
      {
        StringReader stringReader = new StringReader(doc.InnerXml);
        XmlTextReader reader = new XmlTextReader(stringReader);
        XmlValidatingReader validatingReader = new XmlValidatingReader(reader);
        validatingReader.ValidationType = ValidationType.Schema;
        validatingReader.ValidationEventHandler += new ValidationEventHandler(MyValidationEventHandler);
        
        XmlSchemaCollection schemaCollection = new XmlSchemaCollection();
        
        schemaCollection.Add("http://pico.sns.it/pico/1.1/", "http://pico.sns.it/pico/1.1/pico.xsd");
        schemaCollection.Add("xsi", "http://www.w3.org/2001/XMLSchema-instance");
        schemaCollection.Add("dc", "http://purl.org/dc/elements/1.1/");
        schemaCollection.Add("dcterms", "http://purl.org/dc/terms");
        validatingReader.Schemas.Add(schemaCollection);

        /*
            pico     http://pico.sns.it/pico/1.1/
            xsi      http://www.w3.org/2001/XMLSchema-instance
            dc       http://purl.org/dc/elements/1.1/
            dcterms  http://purl.org/dc/terms
            
				    http://pico.sns.it/pico/1.1/ http://pico.sns.it/pico/1.1/pico.xsd
            http://purl.org/dc/elements/1.1/ http://www.dublincore.org/schemas/xmls/qdc/2006/01/06/dc.xsd
            http://purl.org/dc/terms/ http://www.dublincore.org/schemas/xmls/qdc/2006/01/06/dcterms.xsd  
        */
 
        while (validatingReader.Read())
        {
          System.Console.Out.WriteLine(validatingReader.Name);
          // Can add code here to process the content.
        }
        validatingReader.Close();
      }
      catch(Exception e)
      {
        System.Console.Error.WriteLine(e.Message);
        return false;
      }

      return true;
    }

    public static void MyValidationEventHandler(object sender, 
      ValidationEventArgs args) 
    {
      //isValid = false;
      Console.WriteLine("Validation event\n" + args.Message);
    }

    static XmlNode FindNode(XmlNode node, string name)
    {
      XmlNode tmp = null;

      foreach(XmlNode child in node.ChildNodes)
      {
        //System.Console.WriteLine(child.Name);
        
        if(child.Name == name)
        {
          return child;
        }
        
        tmp =  FindNode(child, name);
        if(tmp != null)
          return tmp;
      }

      return tmp;
    }

    static void ProcessHttpError(string url, string info)
    {
      if(_currSet == null)
        _currSet = "NULL";
      
      if(_setStatInfo.ContainsKey(_currSet))
      {
        ((SetStatInfo)_setStatInfo[_currSet]).totHttpErrors++;        
      }

      System.Console.Out.WriteLine("Errore HTTP: {0}", info);
    }

    static void ProcessValidatorError(string info)
    {
      if(_currSet == null)
        _currSet = "NULL";
      
      if(_setStatInfo.ContainsKey(_currSet))
      {
        ((SetStatInfo)_setStatInfo[_currSet]).totValidatorErrors++;
        ((SetStatInfo)_setStatInfo[_currSet]).validatorErrors.Add(info);
      }

      System.Console.Out.WriteLine("Errore di validazione: {0}", info);
    }

    static void ProcessHarvesterError(string reqUri, string info)
    {
      if(_currSet == null)
        _currSet = "NULL";

      if(_setStatInfo.ContainsKey(_currSet))
      {
        ((SetStatInfo)_setStatInfo[_currSet]).totHarvesterErrors++;
        ((SetStatInfo)_setStatInfo[_currSet]).harvesterErrors.Add(string.Format("Request URI: {0}\n{1}",
          reqUri, info));
      }
      
      System.Console.Out.WriteLine("Errore di harvesting: richiesta OAI-PMH: {0}\n{1}\n", reqUri, info);
    }

    static void ProcessMetadata(XmlQualifiedName qname, XmlQualifiedName xsiType, 
      string xmlLang, string text, ArrayList dcsvComponents)
    {      

      string md = qname.Name;
      if(xsiType != null)
        if(xsiType.Name.Length > 0)
          md += (" " + xsiType);
      if(xmlLang.Length > 0)
        md += (" " + xmlLang);

      if(_currSet == null)
        _currSet = "NULL";

      if(_setStatInfo.ContainsKey(_currSet))
      {
        if(((SetStatInfo)_setStatInfo[_currSet]).metadataFreq.ContainsKey(md))
          ((SetStatInfo)_setStatInfo[_currSet]).metadataFreq[md] =
            ((int)((SetStatInfo)_setStatInfo[_currSet]).metadataFreq[md]) + 1;
        else
          ((SetStatInfo)_setStatInfo[_currSet]).metadataFreq[md] = 1;
      }

      /*
      if(_md.ContainsKey(md))
        _md[md] = ((int)_md[md]) + 1;
      else
        _md[md] = 1;
      */

      if(qname.Name.Equals("subject"))
      {
        _subjectCount++;

        if(xsiType.Name.Equals("Thesaurus"))
        {
          string id = "";
          string name = "";

          if(dcsvComponents != null)
          {
            foreach(Validator.DcsvComponent comp in dcsvComponents)
            {
              if(comp._label != null)
              {
                if(comp._label.ToLower().Equals("id"))
                  id = comp._value;
                if(comp._label.ToLower().Equals("name"))
                  name = comp._value;
              }
            }
          }
          if(id.Length > 0 && name.Length > 0)
            CheckThesaurus(id, name);
          else
            CheckThesaurus4(text);
        }
      }
      if(qname.Name.Equals("type"))
      {
        _typeCount++;
      }
      if(qname.Name.Equals("title"))
      {
        _titleCount++;
      }
      if(qname.Name.Equals("identifier"))
      {
        _identifierCount++;
      }
    }

    static bool MyEquals(string s1, string s2)
    {
      if(s1.Length != s2.Length)
        return false;

      for(int n = 0; n < s1.Length; n++)
      {
        if(s1[n] != s2[n])
          return false;
      }

      return true;
    }

    static string RemoveSpaces(string t)
    {
      string tt = t.Trim();

      tt = tt.Replace('\t', ' ');
      tt = tt.Replace('\r', ' ');
      tt = tt.Replace('\n', ' ');
      
      while(tt.IndexOf("  ") >= 0)
      {
        tt = tt.Replace("  ", " ");
      }

      return tt;
    }

    static void CheckThesaurus(string id, string name)
    {
      // name = RemoveSpaces(name);
      
      string xpath = string.Format("//th:relation[th:termId=\"{0}\"]", id);

      XmlNode nodeRelation = _thesaurus.SelectSingleNode(xpath, _thesaurusNamespaceManager);
      if(nodeRelation == null)
      {
        System.Console.Out.WriteLine("Errore: termine non presente nel thesaurus (id='{0}', name='{1}')", 
          id, name);
      }
      else
      {
        XmlNode nodeTermName = nodeRelation.SelectSingleNode("th:termName", _thesaurusNamespaceManager);

        if(!nodeTermName.InnerText.ToLower().Equals((name.ToLower())))
          System.Console.Out.WriteLine("Attenzione: il value del thesaurus non corrisponde (id='{0}', value='{1}')", 
            id, name);
        _thesSubjectCount++;
      }
    }

    static void CheckThesaurus4(string uri)
    {
      // uri = RemoveSpaces(uri);
      
      string xpath = string.Format("/rdf:RDF/rdf:Description[@rdf:about=\"{0}\"]", uri);

      XmlNode nodeTerm = _thesaurus4.SelectSingleNode(xpath, _thesaurus4NamespaceManager);
      if(nodeTerm == null)
      {
        System.Console.Out.WriteLine("Errore: termine non presente nel thesaurus 4 (uri='{0}')", uri);
      }
      else
      {
        _thes4SubjectCount++;
      }
    }

    static void PrintStatInfo()
    {
      System.Console.Out.WriteLine("Statistiche");
      foreach(string setSpec in _setStatInfo.Keys)
      {
        SetStatInfo info = (SetStatInfo)_setStatInfo[setSpec];
        System.Console.Out.WriteLine("Set: \"{0}\"", setSpec);
        if(info.processed)
        {          
          System.Console.Out.WriteLine("Totale identificatori : {0}", info.totIdentifiers);
          System.Console.Out.WriteLine("Record richiesti      : {0}", info.totRecords);
          System.Console.Out.WriteLine("Errori HTTP           : {0}", info.totHttpErrors);
          System.Console.Out.WriteLine("Errori harvester      : {0}", info.totHarvesterErrors);
          foreach(string err in info.harvesterErrors)
          {
            System.Console.Out.WriteLine(err);
          }
          System.Console.Out.WriteLine("Errori validatore     : {0}\n", info.totValidatorErrors);
          foreach(string err in info.validatorErrors)
          {
            System.Console.Out.WriteLine(err);
          }
          System.Console.Out.WriteLine("Metadati:");
          foreach(string md in info.metadataFreq.Keys)
          {
            System.Console.Out.WriteLine("{0}:\t{1}", md, (int)info.metadataFreq[md]);
          }
        }
        else
        {
          System.Console.Out.WriteLine("(non processato)", setSpec);
        }
      }      
    }
	}

  class SetStatInfo
  {
    public bool processed;
    public int errCount;
    public int totIdentifiers;
    public int totRecords;
    public int totHttpErrors;
    public int totHarvesterErrors;
    public int totValidatorErrors;
    //public ArrayList identifiers = null;
    public ArrayList harvesterErrors = null;
    public ArrayList validatorErrors = null;
    public Hashtable metadataFreq = null;

    public SetStatInfo()
    {
      processed = false;
      errCount = 0;
      totIdentifiers = 0;
      totRecords = 0;
      totHttpErrors = 0;
      totValidatorErrors = 0;
      totHarvesterErrors = 0;
      //identifiers = new ArrayList();
      harvesterErrors = new ArrayList();
      validatorErrors = new ArrayList();
      metadataFreq = new Hashtable();
    }    
  }
}
